# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to extract OpenWrt device information."""

from collections.abc import Mapping
import json
from typing import Any

from mobly.controllers.wifi.lib import constants

OpenWrtDevice = Any


def get_device_info(device: 'OpenWrtDevice') -> Mapping[str, str]:
  """Gets OpenWrt device information."""
  device_info = {'serial': device.serial}

  # Get info from the official release info file.
  openwrt_release = device.ssh.execute_command(
      command=f'cat {constants.OPENWRT_RELEASE_INFO_FILE_PATH}',
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
      ignore_error=True,
  )
  match = constants.DEVICE_INFO_PATTERN.fullmatch(openwrt_release)
  if match is not None:
    device_info.update(match.groupdict())

  # Get info from the info file created by the custom image build process.
  device_info.update(_get_info_from_custom_image_build_process(device))
  return device_info


def _get_info_from_custom_image_build_process(
    device: 'OpenWrtDevice',
) -> Mapping[str, str]:
  """Gets device info recorded by the custom image build process."""
  custom_info_str = device.ssh.execute_command(
      command=f'cat {constants.CURSTOM_RELEASE_INFO_FILE_PATH}',
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
      ignore_error=True,
  )
  if not custom_info_str:
    return {}

  custom_info = json.loads(custom_info_str)
  device_info = {}
  device_info['build_profile'] = custom_info.get('standardBuildConfig', {}).get(
      'buildProfile', ''
  )
  device_info['device_name'] = custom_info.get('standardBuildConfig', {}).get(
      'deviceName', ''
  )
  device_info['image_uuid'] = custom_info.get('imageUuid', '')
  device_info['build_time'] = custom_info.get('buildTime', '')
  device_info['custom_image_name'] = custom_info.get('customImageName', '')
  device_info['router_features'] = ','.join(
      custom_info.get('routerFeatures', '')
  )
  return device_info



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Mobly controller module for local sniffer devices connected to lab host."""

from collections.abc import Mapping, Sequence
import logging
import os
from typing import Any

from mobly import logger as mobly_logger
from mobly import runtime_test_info
from mobly import utils

from mobly.controllers.wifi.lib import local_sniffer_manager
from mobly.controllers.wifi.lib import wifi_configs


MOBLY_CONTROLLER_CONFIG_NAME = 'LocalSnifferDevice'
_DEVICE_TAG = MOBLY_CONTROLLER_CONFIG_NAME


class Error(Exception):
  """Error class for the LocalSnifferDevice controller."""


def create(configs: list[dict[str, Any]]) -> list['LocalSnifferDevice']:
  """Creates local sniffer device instances.

  The configs come from Mobly configs that look like:

    ```config.yaml
    TestBeds:
    - Name: SampleTestBed
      Controllers:
        LocalSnifferDevice:
        - phy: 'phy2'
          model: 'ALFA AWUS036AXML'
    ```

  Each config should have the required key-value pair 'phy'. We recommend to
  include model info through key `model`.

  Args:
    configs: a list of dicts, each representing a configuration for a local
      sniffer device.

  Returns:
    A list of LocalSnifferDevice objects.

  Raises:
    Error: Raised if got invalid controller configs.
  """
  if not configs:
    raise Error(f'Missing configuration {configs!r}.')
  devices = [LocalSnifferDevice(config) for config in configs]
  return devices


def destroy(devices: list['LocalSnifferDevice']) -> None:
  """Closes all created local sniffer device instances."""
  for device in devices:
    try:
      device.teardown()
    except Exception:  # pylint: disable=broad-except
      logging.exception('Failed to clean up properly.')


def get_info(
    devices: Sequence['LocalSnifferDevice'],
) -> Sequence[Mapping[str, Any]]:
  """Gets info from the local sniffer device instances used in a test run.

  Args:
    devices: A list of local sniffer device instances.

  Returns:
    A list of dict, each representing info for a device object.
  """
  return [d.device_info for d in devices]


class LocalSnifferDevice:
  """Mobly controller for local sniffer devices connected to lab host.

  Attributes:
    serial: A string that identifies this local sniffer device.
    log_path: A string that is the path where all logs collected on this device
      should be stored.
    log: A logger adapted from root logger with an added prefix specific to a
      remote test machine. The prefix is "[LocalSnifferDevice|<self.serial>] ".
    device_info: A collection of device information.
  """

  def __init__(self, config: dict[str, Any]):
    if 'phy' not in config:
      raise Error(
          f'Missing required field "phy" in device configuration {config!r}.'
      )
    self._phy = config['phy']
    self._model = config.get('model', None)
    self._device_info = None
    self.serial = f'{self._phy}'

    log_path = getattr(logging, 'log_path', '/tmp/logs')
    log_filename = mobly_logger.sanitize_filename(
        f'{_DEVICE_TAG}_{self.serial}'
    )
    self.log_path = os.path.join(log_path, log_filename)
    utils.create_dir(self.log_path)
    self.log = mobly_logger.PrefixLoggerAdapter(
        logging.getLogger(),
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                f'[{_DEVICE_TAG}|{self.serial}]'
            )
        },
    )
    self._sniffer_manager_obj = None

  def __repr__(self) -> str:
    return f'<{_DEVICE_TAG}|{self.serial}>'

  @property
  def device_info(self) -> Mapping[str, str]:
    """Information to be pulled into controller info in the test summary."""
    if self._device_info:
      return self._device_info
    self._device_info = {'phy': self._phy}
    if self._model is not None:
      self._device_info['model'] = self._model
    return self._device_info

  @property
  def _sniffer_manager(self):
    if self._sniffer_manager_obj is None:
      self._sniffer_manager_obj = local_sniffer_manager.LocalSnifferManager(
          self, self._phy
      )
    return self._sniffer_manager_obj

  def start_packet_capture(
      self,
      wifi_config: wifi_configs.WiFiConfig | None = None,
      freq_config: wifi_configs.FreqConfig | None = None,
      capture_config: wifi_configs.PcapConfig | None = None,
  ):
    """Starts pacaket capture on a specific frequency band.

    You need to provide either wifi_config or freq_config. If you provide
    `freq_config`, this will monitor the frequency band specified by it. If you
    provide `wifi_config`, this method will extract channel info from it and
    monitor the same channel that is used by the WiFi network.

    Args:
      wifi_config: The WiFi network to capture the packets.
      freq_config: The frequency to capture the packets.
      capture_config: The configuration to control the packet capture process.

    Raises:
      Error: Raised if the provided configurations are invaliad.
    """
    if (
        sum([
            wifi_config is not None,
            freq_config is not None,
        ])
        != 1
    ):
      raise Error(
          'You must provide either wifi_config or freq_config to start packet'
          f' capture, got wifi_config: "{wifi_config}", freq_config:'
          f' "{freq_config}".'
      )
    if wifi_config is not None:
      freq_config = wifi_configs.get_freq_config(wifi_config)
    self._sniffer_manager.start_capture(
        freq_config=freq_config, capture_config=capture_config
    )

  def stop_packet_capture(
      self, current_test_info: runtime_test_info.RuntimeTestInfo | None = None
  ):
    """Stops pacaket capture."""
    self._sniffer_manager.stop_capture(current_test_info=current_test_info)

  def get_capture_file(self) -> str | None:
    """Gets the full path of the last capture."""
    return self._sniffer_manager.get_capture_file()

  def teardown(self):
    """Tears the device object down."""
    self.log.info('Tearing down this controller object.')
    if self._sniffer_manager_obj is not None:
      self._sniffer_manager_obj.teardown()
      self._sniffer_manager_obj = None



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The util module for generating excerpts based on a host file."""

import io
import pathlib

from mobly import utils


class ClipFileError(Exception):
  """Raised when trying to conduct invalid operation on FileClipper."""


class FileClipper:
  """The util class for clipping a given host file to multiple fragments.

  This receives a file which is continually being written new content. Users can
  call `clip_new_content` multiple times. Each call will clip the newly appended
  content after the last call as a fragment, and write it to a new file.

  The key point of this class is that it maintains the stream position for all
  `clip_new_content` calls. When `clip_new_content` is called, it reads the
  given file by moving the read stream position to the file end. Then it keeps
  this position unchanged, and the next `clip_new_content` call will start
  reading the file from that position. In this way, each `clip_new_content` call
  only contains the newly appended content after the last call.

  Typical usage:

  .. code-block:: python

    # Preparation
    file_clipper = FileClipper('/tmp/log')

    ... # Append 3 lines to /tmp/log and clip these 3 lines to /tmp/log_clip_1
    file_clipper.clip_new_content('/tmp/log_clip_1')

    ... # Append 5 lines to /tmp/log and clip these 5 lines to /tmp/log_clip_2
    file_clipper.clip_new_content('/tmp/log_clip_2')

    ... # Append 2 lines to /tmp/log and clip these 2 lines to /tmp/log_clip_3
    file_clipper.clip_new_content('/tmp/log_clip_3')

    # Close this clipper and it can no longer clip new contents
    file_clipper.close()
  """

  def __init__(
      self,
      source_file_path: pathlib.Path,
      ignore_existing_content: bool = False,
  ) -> None:
    """Initializes an instance and opens the file object to read the given file.

    Args:
      source_file_path: The host path of the source file to clip.
      ignore_existing_content: Whether to ignore the existing content in the
        source file before initializing the clipper. Defaults to False.
    """
    self._source_file_path = source_file_path
    self._file_obj_to_read = None
    self._ignore_existing_content = ignore_existing_content
    # We intentionally hide the `open` method from users and open the file
    # object in the constructor, because no use case shall open the file object
    # multiple times.
    self._open()

  def __del__(self):
    self.close()

  def _open(self) -> None:
    """Opens this clipper.

    This method opens the file object to read the given host file.
    """
    utils.create_dir(str(self._source_file_path.parent))
    self._source_file_path.touch()
    self._file_obj_to_read = io.open(
        self._source_file_path, 'r', encoding='utf-8', errors='replace')
    if self._ignore_existing_content:
      self._file_obj_to_read.seek(0, io.SEEK_END)

  def close(self) -> None:
    """Closes all resources acquired in this instance."""
    if self._file_obj_to_read:
      self._file_obj_to_read.close()
      self._file_obj_to_read = None

  def clip_new_content(self, clip_file_path: pathlib.Path) -> None:
    """Clips the newly appended content and saves it to the given file path.

    Each time this method is called, this will create a clip file based on the
    given source file. The clip contains the content that was written to the
    given source file after the last `clip_new_content` call, or from the start
    of the file.

    Args:
      clip_file_path: The host path to save the created clip.

    Raises:
      ClipFileError: If trying to clip new content after this object is closed.
    """
    if self._file_obj_to_read is None:
      raise ClipFileError(
          'Cannot call `clip_new_content` after this object is closed.')

    with io.open(
        clip_file_path, 'w', encoding='utf-8', errors='replace') as out:
      while True:
        line = self._file_obj_to_read.readline()
        if not line:
          break
        out.write(line)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Configurations for the encryption mode WEP."""

import enum
import random
import string

from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import utils
from mobly.controllers.wifi.lib.encryption import base_encryption_config

# Length for valid WEP keys.
# Length for valid WEP keys in the form of ascii strings.
_ASCII_KEY_STRING_LENGTH = (5, 13, 16)

# Length for valid WEP keys in the form of hex strings.
_HEX_KEY_STRING_LENGTH = (10, 26, 32)


@enum.unique
class AuthAlgs(enum.IntEnum):
  """Enum for the authentication algorithms.

  This corresponds the hostapd configuration field `auth_algs`.
  """

  # Open System Authentication.
  OPEN = 1

  # Shared Key Authentication.
  SHARED = 2


def _gen_random_wep104_key() -> str:
  """Generates a random 104-bit WEP key in the form of hex string."""
  # length: 104 bits / 4 bits per hex digit = 26
  return ''.join(random.choice(string.hexdigits) for _ in range(26))


class Wep(base_encryption_config.BaseEncryptionConfig):
  """The configuration class for the WEP encryption type."""

  _auth_algs: AuthAlgs
  _keys: list[str]
  _default_key: int
  _password: str | None = None

  def __init__(
      self,
      auth_algs: AuthAlgs = AuthAlgs.OPEN,
      keys: list[str] | None = None,
      default_key: int = 0,
  ):
    """Constructor.

    If no key is speicified, a random 104-bit key will be used.

    Args:
      auth_algs: The authentication algorithm to use.
      keys: All WEP keys to use, which should be at most 4 keys.
      default_key: The key index to use. The corresponding key must be provided
        in the argument `keys`.
    """
    self._auth_algs = auth_algs
    self._default_key = default_key
    if default_key == 0 and keys is None:
      # Generate a wep104 key as default.
      self._keys = [_gen_random_wep104_key()]
    else:
      self._keys = keys or []
    self._password = None
    self._validate()

  def _validate(self):
    """Validates the configurations are valid."""
    if len(self._keys) > 4:
      raise errors.ConfigError(
          'At most 4 keys can be set for WEP encryption type, got'
          f' {len(self._keys)} keys.'
      )

    if self._default_key >= len(self._keys) or self._default_key < 0:
      raise errors.ConfigError(
          f'Default key index {self._default_key} out of range'
          f' [0, {len(self._keys) - 1}].'
      )

    for key in self._keys:
      self._validate_key(key)

    if self._auth_algs == AuthAlgs.SHARED:
      self._password = self._keys[self._default_key]
    else:
      self._password = None

  def _validate_key(self, key: str):
    if len(key) in _HEX_KEY_STRING_LENGTH:
      if not utils.is_hex_string(key):
        raise errors.ConfigError(
            'WEP encryption key with length 10, 26, or 32 should only contain'
            f' hexadecimal digits: {key}'
        )
      return

    if len(key) in _ASCII_KEY_STRING_LENGTH:
      if not key.isascii():
        raise errors.ConfigError(
            f'WEP encryption key with length {len(key)} should only contain'
            f' ascii chars: {key}'
        )
      return

    raise errors.ConfigError(
        f'Got invalid WEP encryption key length: {len(key)}'
    )

  def update_hostapd_conf(self, hostapd_conf):
    """See docstring of base class."""
    hostapd_conf.set_password(self._password)
    hostapd_conf.update('auth_algs', str(self._auth_algs.value))
    hostapd_conf.update('wep_default_key', str(self._default_key))
    for i, key in enumerate(self._keys):
      hostapd_conf.update(f'wep_key{i}', self._format_key(key))

  def _format_key(self, key: str) -> str:
    """Formats keys according to the hostapd configuration requirements."""
    if len(key) in _ASCII_KEY_STRING_LENGTH:
      # When using the ASCII string format, keys must be surrounded by quotes.
      return f'"{key}"'
    return key



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The module for streaming the output of a ChannelFile object to host."""

from concurrent import futures
import io
import logging
import os
import socket
import threading

from mobly import utils
from paramiko import channel


# The default timeout seconds for each channel read in RemotePopen
CHANNEL_READ_DEFAULT_TIMEOUT_SEC = 5

# The default timeout seconds for ChannelFileStreamer to stop
CHANNEL_FILE_STREAMER_STOP_TIMEOUT_SEC = 10


class ChannelFileStreamer:
  """The class for streaming the output of a ChannelFile object to a host file.

  This class will create a thread to continuously read data from the given
  channel file object and append to a host file.
  """

  def __init__(self, channel_file_obj: channel.ChannelFile,
               output_file_path: str, logger: logging.LoggerAdapter) -> None:
    """Initializes the ChannelFileStreamer instance.

    Args:
      channel_file_obj: The channel file object to read data.
      output_file_path: The host file path for appending the data read from the
        channel file object.
      logger: The logger.
    """
    self._channel_file_obj = channel_file_obj
    self._output_file_path = output_file_path
    utils.create_dir(os.path.dirname(output_file_path))
    self._output_file_obj = io.open(self._output_file_path, 'a')
    self._log = logger

    # When streaming output, the channel timeout is the timeout argument for
    # each channel read. We set it to a small value to avoid being blocked
    # indefinitely.
    self._channel_file_obj.channel.settimeout(CHANNEL_READ_DEFAULT_TIMEOUT_SEC)

    self._lock = threading.Lock()
    self._executor = futures.ThreadPoolExecutor(
        max_workers=1, thread_name_prefix='ChannelFileStreamer')
    self._job = None

  def _stream_channel_file_output(self, channel_file_obj: channel.ChannelFile,
                                  output_file_obj: io.TextIOWrapper,
                                  lock: threading.Lock,
                                  logger: logging.LoggerAdapter) -> None:
    """Streams the channel file output to a host file object."""
    logger.debug('Starting to stream the output to host file %s',
                 output_file_obj.name)

    while True:
      with lock:
        if output_file_obj.closed:
          logger.debug(
              'Stopping log streaming because the target output file object '
              'is already closed.')
          break

      if channel_file_obj.closed:
        logger.debug(
            'Stopping log streaming because the input channel file object '
            'is already closed.')
        break

      try:
        output = channel_file_obj.readline()

        if not output:
          logger.debug(
              'Stopping log streaming because got EOF from the input channel '
              'file object.')
          break

        try:
          output = output.decode('utf-8')
        except UnicodeDecodeError:
          logger.error('Ignoring the channel file output decoding error. The '
                       'output failed to decode is: "%s"', output)
          continue

        with lock:
          if output_file_obj.closed:
            logger.debug(
                'Stopping log streaming because the target output file '
                'object is already closed. Ignoring the unwritten log: %s',
                output)
            break

          output_file_obj.write(output)
          output_file_obj.flush()

      except socket.timeout:
        logger.debug('Ignoring the timeout error when waiting for the output.')

    logger.debug('Finished streaming the output to host file %s',
                 output_file_obj.name)

  def start(self) -> None:
    """Starts streaming the channel file output."""
    self._job = self._executor.submit(self._stream_channel_file_output,
                                      self._channel_file_obj,
                                      self._output_file_obj, self._lock,
                                      self._log)

  def stop(self,
           timeout_sec: int = CHANNEL_FILE_STREAMER_STOP_TIMEOUT_SEC) -> None:
    """Stops streaming the channel file output.

    If the streaming thread exited with an exception, this will raise that
    exception in the main thread.

    Args:
      timeout_sec: The timeout seconds to wait for the streaming thread to exit.
        If the thread does not exit within the given time, we will force stop
        it, potentially causing subsequent outputs to be discarded.
    """
    if timeout_sec < self._channel_file_obj.channel.timeout:
      raise ValueError(
          'The stop timeout should be no smaller than the channel '
          f'read timeout, got stop timeout={timeout_sec}, channel '
          f'read timeout={self._channel_file_obj.channel.timeout}'
      )

    if self._job is None:
      raise ValueError('job not initialized')
    try:
      self._job.result(timeout=timeout_sec)
    except futures.TimeoutError:
      self._log.debug(
          'Ignoring the timeout error occurred when trying to stop,'
          f' which may cause the log file {self._output_file_path}'
          ' incomplete.'
      )
      with self._lock:
        self._output_file_obj.close()

    self._log.debug('Shutting down the threading pool executor.')
    self._executor.shutdown()

    if (exception := self._job.exception()) is not None:
      raise exception



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities to wrap around running the `iw` prgoram on AP devices.

See following file for an example of the `iw phy` output:
//testing/mobly/platforms/wifi/test_data/iw_phy_output.txt
"""

from collections.abc import Sequence
import dataclasses
import enum
import re
from typing import Any, Self

from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors

OpenWrtDevice = Any

_FREQUENCY_INFO_RE = re.compile(
    r'.*\* \d+(\.\d)? MHz \[(?P<channel>\d+)\](?: \([0-9.]+ dBm\))?(?:'
    r' \((?P<flags>[a-zA-Z, ]+)\))?.*'
)
_FREQUENCY_INFO_RE_GROUP_CHANNEL = 'channel'
_FREQUENCY_INFO_RE_GROUP_FLAGS = 'flags'

_BAND_RE = re.compile(r'Band (?P<num>\d+):')
_BAND_RE_GROUP_NUM = 'num'

_PHY_NAME_RE = re.compile(r'Wiphy (?P<name>\S+)')
_PHY_NAME_RE_GROUP_NAME = 'name'

_PHY_INDEX_RE = re.compile(r'wiphy index: (?P<phyindex>\d+)')
_PHY_INDEX_RE_GROUP_PHYINDEX = 'phyindex'

_PREFIX_TAB_CHARACTERS_RE = re.compile(r'^\t+')

_TEXT_LABEL_FREQUENCIES = 'Frequencies:'

_STATION_TITLE_RE = re.compile(
    r'^Station (?P<mac_address>[:0-9a-fA-F]+) \(on \S+\)'
)
_STATION_TITLE_RE_GROUP = 'mac_address'

_ASSOCIATED_AT_RE = re.compile(r'associated at:\s*(?P<timestamp>\d+) ms')
_ASSOCIATED_AT_RE_GROUP = 'timestamp'

_IW_DEV_RE = re.compile(
    r'Interface\s+(?P<name>.*)(.|\n)*?'
    r'addr\s+(?P<addr>.*)(.|\n)*?'
    r'(ssid\s+(?P<ssid>.*)(.|\n)*?)?'
    r'type\s+(?P<type>.*)'
)


class IwOutputParsingError(errors.BaseError):
  """Failed to parse the output of the `iw` command."""


class NoSuchStationError(errors.BaseError):
  """Failed to find the specified station."""


@enum.unique
class ChannelFlags(enum.StrEnum):
  """Channel flags in the `iw phy` output."""

  RADAR_DETECTION = 'radar detection'


@dataclasses.dataclass(frozen=True, eq=True, order=True)
class Channel:
  """Channel information in the `Frequencies` section of `iw phy` output."""

  num: int
  flags: Sequence[str]


@dataclasses.dataclass(frozen=True, eq=True, order=True)
class Band:
  """Band information in the `Bands` section of `iw phy` output."""

  num: int
  channels: Sequence[Channel]


@dataclasses.dataclass(frozen=True)
class Interface:
  """Interface information in the `iw dev` output.

  Attributes:
    name: The name of the interface.
    addr: The MAC address of the interface.
    type: Specify the operating mode of the interface.
    ssid: The SSID of the interface, None if the interface is not a broadcasting
      Wi-Fi.
  """

  name: str
  addr: str
  type: str
  ssid: str | None = None


@dataclasses.dataclass(frozen=True, eq=True, order=True)
class Phy:
  """Wireless hardware device information in the `iw phy` output."""

  name: str
  phyindex: int
  bands: Sequence[Band]


@dataclasses.dataclass(frozen=True)
class Station:
  """Station information in the `iw dev {devname} station dump/get` output.

  Attributes:
    mac_address: The MAC address of the station.
    associated_at_timestamp_ms: The timestamp when the station associated with
      the AP.
  """

  mac_address: str
  associated_at_timestamp_ms: int


@dataclasses.dataclass(frozen=True, eq=True, order=True)
class _TreeNode:
  """The class for representing `iw` output into Tree structure.

  See docstring of `parse_indented_text` for parse logic.
  """

  text: str | None
  indent_level: int
  children: list[Self] = dataclasses.field(default_factory=list)

  def find_child_by_regex(
      self, *, pattern: re.Pattern[str]
  ) -> tuple[Self, re.Match[str]]:
    """Finds the direct child of which the text matches the given pattern."""
    matched_children = []
    for c in self.children:
      if (match := pattern.match(c.text)) is not None:
        matched_children.append((c, match))
    if not matched_children:
      raise IwOutputParsingError(
          f'Did not find child with pattern: "{pattern}"'
      )
    if len(matched_children) > 1:
      raise IwOutputParsingError(
          f'Found multiple children with pattern: "{pattern}"'
      )
    return matched_children[0]

  def find_children_by_regex(
      self, *, pattern: re.Pattern[str]
  ) -> Sequence[Self]:
    """Finds all direct children of which the text match the regex pattern."""
    return [c for c in self.children if pattern.match(c.text)]

  def find_child(self, *, text: str) -> Self:
    """Finds the direct child of which the text equals the given text."""
    matched_children = [c for c in self.children if text == c.text]
    if not matched_children:
      raise IwOutputParsingError(f'Did not find child with text: "{text}"')
    if len(matched_children) > 1:
      raise IwOutputParsingError(f'Found multiple children with text: "{text}"')
    return matched_children[0]


def _get_indent_level(line: str) -> int:
  r"""Returns the indent level of the given line.

  The indent level means the number of `\t` characters at the start of this
  line.

  Args:
    line: The line to calculate indent level.

  Returns:
    The indent level.
  """
  prefix = _PREFIX_TAB_CHARACTERS_RE.match(line)
  if prefix is None:
    return 0
  return len(prefix[0])


def _recursively_parse_to_tree(
    lines: Sequence[str], parent_node: _TreeNode
) -> Sequence[str]:
  """Recursively parses the lines into a Tree.

  This function tries to parse lines `lines` to nodes in the subtree of node
  `parent_node`. `lines` may contain lines that are out of the `parent_node`
  subtree, this will return those lines.

  Args:
    lines: All the lines to be parsed.
    parent_node: The node that is considered the root node of the subtree.

  Returns:
    Return the lines that are out of the scope of `parent_node` subtree.

  Raises:
    IwOutputParsingError: Failed to parse the lines.
  """
  while lines:
    line = lines[0]
    if not line:
      lines = lines[1:]
      continue

    current_indent_level = _get_indent_level(line)

    line = line.strip()
    node = _TreeNode(text=line, indent_level=current_indent_level)

    if current_indent_level > parent_node.indent_level + 2:
      raise IwOutputParsingError(
          'Indented text parse error with a larger than 2 indent level'
          f' difference. Line: "{line}", parent_node: "{parent_node}"'
      )

    if current_indent_level == parent_node.indent_level + 1:
      parent_node.children.append(node)
      lines = lines[1:]
      continue

    if current_indent_level == parent_node.indent_level + 2:
      # This clause means current line is a child node of the last parsed line,
      # so we need to call _recursively_parse_to_tree to parse this line, with
      # last node as parent_node.
      if not parent_node.children:
        raise IwOutputParsingError(
            f'Indented text parse error with line "{line}", parent_node'
            f' "{parent_node}"'
        )

      child_node = parent_node.children[-1]
      child_node.children.append(node)
      lines = _recursively_parse_to_tree(lines[1:], child_node)
      continue

    # This clause means current line is not within the subtree of the
    # `parent_node`, so directly return.
    return lines

  return lines


def parse_indented_text(text: str) -> _TreeNode:
  r"""Parses the indented text into a Tree.

  Following is the example of indented text to parse:
  ```
  Wiphy phy1
  \twiphy index: 1
  \tCapabilities: 0x9ff
  \t\tRX LDPC
  \t\tHT20
  ```

  The structure of the text is defined by the amount of `\t` characters at the
  start of each line. So we will parse the text into the following tree:

  ```
  _TreeNode(    # root
      text=None,
      indent_level=-1,
      children=[
        _TreeNode(
            text='Wiphy phy1',
            indent_level=0,
            children=[
                _TreeNode(
                    text='wiphy indext: 1',
                    indent_level=1,
                    children=[],
                ),
                _TreeNode(
                    text='Capabilities: 0x9ff',
                    indent_level=1,
                    children=[
                        _TreeNode(text='RX LDPC', indent_level=2, children=[]),
                        _TreeNode(text='HT20', indent_level=2, children=[]),
                    ],
                ),
            ],
        ),
      ],
  )
  ```
  Args:
    text: The indented text to parse.

  Returns:
    The root node of the parsed Tree.
  """
  root_node = _TreeNode(text=None, indent_level=-1)
  _recursively_parse_to_tree(
      lines=text.strip().splitlines(), parent_node=root_node
  )
  return root_node


def _parse_tree_to_channels(node: _TreeNode) -> Sequence[Channel]:
  """Parses the tree root at the given `node` to be channels."""
  channels = []
  for child in node.children:
    match = _FREQUENCY_INFO_RE.match(child.text)
    if match is None:
      continue
    channel = int(match.group(_FREQUENCY_INFO_RE_GROUP_CHANNEL))
    flags_match = match.group(_FREQUENCY_INFO_RE_GROUP_FLAGS)
    flags = []
    if flags_match is not None:
      flags = list(map(lambda s: s.strip(), flags_match.split(',')))
    channels.append(Channel(num=channel, flags=flags))
  return channels


def _parse_tree_to_band(node: _TreeNode) -> Band:
  """Parses the tree root at the given `node` to be a band."""
  match = _BAND_RE.match(node.text)
  if match is None:
    raise IwOutputParsingError(
        f'Did not find Band number in node: "{node.text}"'
    )
  num = int(match.group(_BAND_RE_GROUP_NUM))
  channels = _parse_tree_to_channels(
      node.find_child(text=_TEXT_LABEL_FREQUENCIES)
  )
  return Band(num=num, channels=channels)


def get_all_phys(device: 'OpenWrtDevice') -> Sequence[Phy]:
  """Gets all the wireless hardware devices on the given AP device."""
  cmd = constants.Commands.IW_PHY
  output = device.ssh.execute_command(
      command=cmd,
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
  )
  device.log.debug('Parsing command "%s" output to a tree.', cmd)
  tree_root = parse_indented_text(output)

  device.log.debug('Parsing the tree to structured data types.')
  phys = []
  for node in tree_root.children:
    phy_match = _PHY_NAME_RE.match(node.text)
    if phy_match is None:
      raise IwOutputParsingError(
          f'Did not find Wiphy name in node: "{node.text}"'
      )
    phy_name = phy_match.group(_PHY_NAME_RE_GROUP_NAME)
    _, phyindex_match = node.find_child_by_regex(pattern=_PHY_INDEX_RE)
    phyindex = int(phyindex_match.group(_PHY_INDEX_RE_GROUP_PHYINDEX))

    bands = []
    for node in node.find_children_by_regex(pattern=_BAND_RE):
      bands.append(_parse_tree_to_band(node))
    phys.append(Phy(name=phy_name, phyindex=phyindex, bands=bands))
  return phys


def get_all_dfs_channels(phy: Phy) -> set[int]:
  """Gets all DFS channels of the given wireless hardware device."""
  dfs_channels = set()
  for band in phy.bands:
    for channel in band.channels:
      if ChannelFlags.RADAR_DETECTION in channel.flags:
        dfs_channels.add(channel.num)
  return dfs_channels


def get_phy_by_channel(phys: Sequence[Phy], channel: int) -> Phy:
  """Gets the wireless hardware device that can supports the given channel."""
  for phy in phys:
    for band in phy.bands:
      for c in band.channels:
        if c.num == channel:
          return phy
  raise IwOutputParsingError(
      'Did not find a wireless hardware device that supports channel'
      f' "{channel}". All devices: {phys}'
  )


def _parse_iw_station_output(output: str) -> Sequence[Station]:
  """Parses the iw output related to station information.

  Example output of iw station dump/get:

  ```
  Station ae:d9:f5:47:9a:c0 (on managed0)
      inactive time: 36 ms
      rx bytes: 20164
      ...
      connected time: 8 seconds
      associated at [boottime]: 42.448s
      associated at: 1713791172733 ms
      current time: 1713791180538 ms
  ```

  Args:
    output: The iw output.

  Returns:
    A sequence of station.

  Raises:
    IwOutputParsingError: if failed to parse the output.
  """
  tree_root = parse_indented_text(output)
  results = []
  for node in tree_root.children:
    station_title_match = _STATION_TITLE_RE.match(node.text)
    if station_title_match is None:
      raise IwOutputParsingError(
          f'Did not find Station mac address in node: "{node.text}"'
      )
    station_mac_address = station_title_match.group(_STATION_TITLE_RE_GROUP)

    _, match = node.find_child_by_regex(pattern=_ASSOCIATED_AT_RE)
    timestamp = int(match.group(_ASSOCIATED_AT_RE_GROUP))

    results.append(
        Station(
            mac_address=station_mac_address,
            associated_at_timestamp_ms=timestamp,
        )
    )
  return results


def get_all_known_stations(
    device: 'OpenWrtDevice', interface: str
) -> Sequence[Station]:
  """Gets all the known stations on the given wireless interface."""
  cmd = constants.Commands.IW_DEV_STATION_DUMP.format(interface=interface)
  output = device.ssh.execute_command(
      command=cmd,
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
  )
  return _parse_iw_station_output(output)


def get_station_info(
    device: 'OpenWrtDevice', interface: str, mac_address: str
) -> Station:
  """Gets info for the station with the MAC address on the given interface.

  Args:
    device: The AP device controller.
    interface: The interface that the station is associated with.
    mac_address: The MAC address of the station.

  Returns:
    The station information.

  Raises:
    NoSuchStationError: Failed to find the specified station.
  """
  cmd = constants.Commands.IW_DEV_STATION_GET.format(
      interface=interface, station_mac_address=mac_address
  )
  output = device.ssh.execute_command(
      command=cmd,
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
      ignore_error=True,
  )
  if not output:
    raise NoSuchStationError(
        f'Did not find station with mac address "{mac_address}". Got iw command'
        f' output: {output}'
    )
  stations = _parse_iw_station_output(output)
  # The output of the above command will contain one station at most.
  return stations[0]


def get_all_interfaces(device: 'OpenWrtDevice') -> Sequence[Interface]:
  """Gets all interfaces on the given AP device."""
  output = device.ssh.execute_command(
      command=constants.Commands.IW_DEV,
      timeout=constants.CMD_SHORT_TIMEOUT.total_seconds(),
  )
  return [
      Interface(**match.groupdict()) for match in _IW_DEV_RE.finditer(output)
  ]



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The base class for sniffer manager."""

import abc
from typing import Any

from mobly import logger as mobly_logger
from mobly import runtime_test_info

from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import wifi_configs


def gen_iw_set_freq_cmd(
    interface: str,
    freq_config: wifi_configs.FreqConfig,
) -> str:
  """Generates the command to set the frequency of the interface.

  There are 2 formats to call an `iw` command to set frequency:

    set freq <freq> [NOHT|HT20|HT40+|HT40-|5MHz|10MHz|80MHz]
    set freq <control freq> [5|10|20|40|80|80+80|160] [<center1_freq>
      [<center2_freq>]]

  We use the second pattern for channel widht >= 80MHz, and the first one for
  the rest as iw will derive the center frequency for us so we don't have to
  duplicate the logic.

  This function does not validate the given frequencies here and delegate it
  to `iw`. `iw` fails with invalid argument error if the frequencies are not
  matched or invalid.

  Args:
    interface: the wireless interface to set frequency.
    freq_config: the frequency configuration.

  Returns:
    The command to set the frequency.
  """
  freq = constants.CHANNEL_TO_FREQUENCY[freq_config.channel]
  args = [str(freq)]
  match freq_config.ht_mode:
    case (
        wifi_configs.HTMode.NOHT
        | wifi_configs.HTMode.HT20
        | wifi_configs.HTMode.HT40_PLUS
        | wifi_configs.HTMode.HT40_MINUS
    ):
      args.append(str(freq_config.ht_mode))
    case wifi_configs.HTMode.HT80:
      args.append('80')
      args.append(str(freq_config.center1_freq))
  return constants.Commands.IW_DEV_SET_FREQ.format(
      interface=interface, freq_args=' '.join(args)
  )


class SnifferManagerBase(abc.ABC):
  """A base class for managing sniffer process."""

  def __init__(self, device: Any):
    self._device = device
    self._log = mobly_logger.PrefixLoggerAdapter(
        device.log,
        {
            mobly_logger.PrefixLoggerAdapter.EXTRA_KEY_LOG_PREFIX: (
                '[SnifferManager]'
            )
        },
    )

  @property
  def device(self) -> Any:
    """The device controller object."""
    return self._device

  @property
  def log(self) -> mobly_logger.PrefixLoggerAdapter:
    """The logger for this object."""
    return self._log

  @property
  @abc.abstractmethod
  def is_alive(self) -> bool:
    """True if the service is alive; False otherwise."""

  @abc.abstractmethod
  def get_capture_file(self) -> str | None:
    """Gets the full path of the last capture file."""

  def _assert_not_running(self):
    """Asserts this manager is not running.

    Raises:
      errors.SnifferManagerError: if this manager is running.
    """
    if self.is_alive:
      raise errors.SnifferManagerError(
          'Running multiple sniffer instances is not allowed.'
      )

  def start_capture(
      self,
      freq_config: wifi_configs.FreqConfig,
      capture_config: wifi_configs.PcapConfig | None = None,
  ):
    """Starts pacaket capture on the given frequency band."""
    capture_config = capture_config or wifi_configs.PcapConfig()
    self._log.debug(
        'Starting packet capture on frequency conf: %s, capture conf: %s',
        freq_config,
        capture_config,
    )
    self._assert_not_running()
    self.start_sniffer_process(freq_config, capture_config)
    self._log.debug('Started packet capturing.')

  @abc.abstractmethod
  def start_sniffer_process(
      self,
      freq_config: wifi_configs.FreqConfig,
      capture_config: wifi_configs.PcapConfig,
  ) -> None:
    """Starts the sniffer process."""

  def stop_capture(
      self, current_test_info: runtime_test_info.RuntimeTestInfo | None = None
  ):
    """Stops pacaket capture.

    Args:
      current_test_info: If provided, this will move the captured packets to
        `current_test_info.output_path`. Otherwise the captured packets will be
        removed.
    """
    if not self.is_alive:
      return
    self._log.debug('Stopping packet capturing.')
    self.stop_sniffer_process()
    self.post_process_packet_files(current_test_info)
    self._log.debug('Stopped packet capturing.')

  @abc.abstractmethod
  def stop_sniffer_process(self):
    """Stops the sniffer process."""

  @abc.abstractmethod
  def post_process_packet_files(
      self, current_test_info: runtime_test_info.RuntimeTestInfo | None = None
  ):
    """Processes packet files after stopping the sniffer process."""

  def teardown(self):
    """Tears the device object down."""
    self._log.info('Tearing down...')
    self.stop_capture(current_test_info=None)

  def __del__(self):
    self.teardown()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for using iperf to measure network speed."""

from collections.abc import Sequence
import time

from mobly.controllers import android_device

from mobly.controllers.wifi import openwrt_device
from mobly.controllers.wifi.lib import wifi_configs

_IPERF_CMD = 'iperf3'
_DEFAULT_IPERF_TEST_DURATION_SEC = 10

INVALID_INT = -1
DEFAULT_CLIENT_ARGS = '-t {test_duration_sec} -P1'
DEFAULT_SERVER_ARGS = '-f m'
IPERF_SERVER_START_DELAY_SEC = 1


class IperfServerOnOpenWrt:
  """Class that handles iperf3 server operations on OpenWrt devices."""

  def __init__(
      self, device: openwrt_device.OpenWrtDevice, arg: str = DEFAULT_SERVER_ARGS
  ):
    """Constructor."""
    self._device = device
    self._cmd = f'{_IPERF_CMD} -s {arg}'
    self._iperf_process = None

  def start(self):
    """Starts iperf server on the OpenWrt device."""
    if self._iperf_process is not None:
      return

    self._iperf_process = self._device.ssh.start_remote_process(
        command=self._cmd,
        get_pty=True,
    )

  def stop(self):
    """Stops the iperf server on the OpenWrt device."""
    if not self._iperf_process:
      return
    proc = self._iperf_process
    self._iperf_process = None
    proc.terminate(assert_process_exit=True)
    stdout, stderr = proc.communicate()
    self._device.log.info('Iperf server stdout: %s, stderr: %s', stdout, stderr)


def run_iperf_test(
    network_client: android_device.AndroidDevice,
    network_owner: openwrt_device.OpenWrtDevice,
    wifi_info: wifi_configs.WifiInfo,
    test_duration_sec: int = _DEFAULT_IPERF_TEST_DURATION_SEC,
) -> tuple[bool, Sequence[str]]:
  """Runs iperf test from network_client to network_owner.

  Args:
    network_client: Android device that is the client in the iperf test.
    network_owner: OpenWrt device that is the server in the iperf test.
    wifi_info: The information for the WiFi network to be tested.
    test_duration_sec: The duration of the iperf test.

  Returns:
    A tuple of following values:
      status: True if iperf ran successfully.
      results: The iperf data flow information printed by iperf client.
  """
  owner_addr = get_ap_ip(network_owner, wifi_info.interface)
  if not owner_addr:
    return False, []

  client_arg = DEFAULT_CLIENT_ARGS.format(test_duration_sec=test_duration_sec)
  server_arg = DEFAULT_SERVER_ARGS

  server = IperfServerOnOpenWrt(network_owner, server_arg)
  try:
    network_owner.log.info('Starting iperf server.')
    server.start()
    time.sleep(IPERF_SERVER_START_DELAY_SEC)
    network_client.log.info(f'Starting iperf client {owner_addr}.')
    success, result_list = network_client.run_iperf_client(
        owner_addr, client_arg
    )
    return success, result_list
  except android_device.adb.AdbError:
    network_client.log.info('iperf failed on client.')
    network_owner.ssh.execute_command('ifconfig')
    network_client.adb.shell('ifconfig')
  finally:
    server.stop()
  return False, []


def get_ap_ip(
    ap: openwrt_device.OpenWrtDevice,
    interface: str,
) -> str:
  """Gets the IP address of the OpenWrt device."""
  ifconfig = get_ap_ifconfig(ap, interface)
  return get_substr_between_prefix_postfix(ifconfig, 'inet addr:', 'Bcast')


def get_substr_between_prefix_postfix(
    string: str, prefix: str, postfix: str
) -> str:
  """Gets substring between prefix and postfix."""
  right_index = string.rfind(postfix)
  if right_index == -1:
    return ''
  left_index = string[:right_index].rfind(prefix)
  if left_index > 0:
    try:
      return string[left_index + len(prefix) : right_index].strip()
    except IndexError:
      return ''
  return ''


def get_ap_ifconfig(ap: openwrt_device.OpenWrtDevice, interface: str) -> str:
  """Gets network info on the OpenWrt device."""
  return ap.ssh.execute_command(
      f'ifconfig | grep -A6 {interface}', ignore_error=True
  )


def get_ad_ifconfig(ad: android_device.AndroidDevice) -> str:
  """Gets network info from adb shell ifconfig."""
  return ad.adb.shell('ifconfig').decode('utf-8').strip()



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A module for managing sniffer on a local device connected to host."""

import os
import shutil
from typing import Any

from mobly import logger as mobly_logger
from mobly import runtime_test_info
from mobly import utils

from mobly.controllers.wifi.lib import constants
from mobly.controllers.wifi.lib import errors
from mobly.controllers.wifi.lib import sniffer_manager_base
from mobly.controllers.wifi.lib import utils as wifi_utils
from mobly.controllers.wifi.lib import wifi_configs


SnifferDevice = Any

_INTERFACE_PREFIX = 'monitor0'
_FILE_TAG = 'packets'
_FILE_TAG_MERGED = 'packets,merged'

# Capture process adopts rotation with file size limit 50MB and at most 2 files
# for saving captured packets. The packet capture process alternatively saves to
# 2 files. Once it reaches the file size limit it will start overwrite the other
# file.
_PCAP_FILE_SIZE_MB = 300
_KEEP_PCAP_FILE_NUM = 2
_KEEP_LATEST_PCAP_ARG = f'-W {_KEEP_PCAP_FILE_NUM} -C {_PCAP_FILE_SIZE_MB}'

# `tcpdump` filter to ignore QoS data frames.
_FILTER_IGNORE_QOS_DATA_FRAMES = r'not \(type data subtype qos-data\)'


class LocalSnifferManager(sniffer_manager_base.SnifferManagerBase):
  """A class for managing sniffer on a local device connected to host."""

  def __init__(self, device: 'SnifferDevice', phy: str):
    super().__init__(device)
    self._phy = phy
    # We should concact `self._phy` into the interface name because many
    # local sniffer might connect with the same host and used by different
    # tests concurrently.
    self._interface = f'{_INTERFACE_PREFIX}_{self._phy}'
    self._packet_dir = os.path.join(self.device.log_path, _FILE_TAG)
    self._proc = None
    self._timestamp = None
    self._sniffer_stdout = None
    self._sniffer_stderr = None

  @property
  def is_alive(self) -> bool:
    """True if the service is alive; False otherwise."""
    return self._proc is not None

  def get_capture_file(self) -> str | None:
    """Gets the full path of the last capture."""
    return self._capture_file_path

  def start_sniffer_process(
      self,
      freq_config: wifi_configs.FreqConfig,
      capture_config: wifi_configs.PcapConfig,
  ) -> None:
    """Configures network interface and starts the tcpdump process."""
    interface = self._interface
    phy = self._phy

    wifi_utils.run_command(
        cmd=constants.Commands.IW_DEV_DEL.format(interface=interface),
        ignore_error=True,
    )
    wifi_utils.run_command(
        cmd=constants.Commands.IW_DEV_ADD_MONITOR.format(
            phy=phy, interface=interface
        ),
    )
    # TODO: Using sudo permission is a temporary workaround, we
    # should figure out how to allowing running ip link up command. Otherwise
    # we need to modify lab host configuration.
    wifi_utils.run_command(
        cmd=f'sudo {constants.Commands.IP_LINK_UP.format(interface=interface)}',
    )
    wifi_utils.run_command(
        cmd=sniffer_manager_base.gen_iw_set_freq_cmd(interface, freq_config),
    )

    self._remove_capture_dir()
    utils.create_dir(self._packet_dir)
    self._timestamp = mobly_logger.get_log_file_timestamp()
    self._packet_streaming_path = self._get_packet_streaming_path()
    self._sniffer_stdout = open(
        self._get_sniffer_log_path(log_level='stdout'), 'w'
    )
    self._sniffer_stderr = open(
        self._get_sniffer_log_path(log_level='stderr'), 'w'
    )

    capture_args = []
    if capture_config.keep_latest_packets:
      capture_args.append(_KEEP_LATEST_PCAP_ARG)
    if capture_config.ignore_qos_data_frames:
      capture_args.append(_FILTER_IGNORE_QOS_DATA_FRAMES)
    cmd = constants.Commands.START_TCPDUMP.format(
        interface=interface,
        file_path=f'"{self._packet_streaming_path}"',
        args=' '.join(capture_args),
    )
    self._proc = utils.start_standing_subprocess(
        cmd=cmd,
        shell=True,
        stdout=self._sniffer_stdout,
        stderr=self._sniffer_stderr,
    )
    self._print_interface_state()

  def _print_interface_state(self):
    """Prints the state of network interface in DEBUG logs."""
    wifi_utils.run_command(
        cmd=constants.Commands.IW_DEV_INFO.format(interface=self._interface),
    )
    wifi_utils.run_command(
        cmd=constants.Commands.IP_LINK_SHOW.format(interface=self._interface),
    )

  def _remove_capture_dir(self):
    """Removes the directory if it exists."""
    if os.path.exists(self._packet_dir):
      shutil.rmtree(self._packet_dir)

  def stop_sniffer_process(self):
    """Stops the sniffer process and closes log file objects."""
    if (proc := self._proc) is None:
      return

    self._proc = None
    if proc.poll() is not None:
      self._try_close_file_objects()
      raise errors.SnifferManagerError(
          f'Sniffer process (pid={proc.pid}) already stopped.'
      )

    try:
      utils.stop_standing_subprocess(proc)
    finally:
      self._try_close_file_objects()
    self.log.debug(
        'Stopped the sniffer process (pid=%d) with output: %s',
        proc.pid,
        proc.communicate(),
    )

  def _try_close_file_objects(self):
    """Closes log file objects."""
    if self._sniffer_stdout is not None:
      self._sniffer_stdout.close()
      self._sniffer_stdout = None
    if self._sniffer_stderr is not None:
      self._sniffer_stderr.close()
      self._sniffer_stderr = None

  def post_process_packet_files(
      self, current_test_info: runtime_test_info.RuntimeTestInfo | None = None
  ):
    """Processes packet files after stopping the sniffer process."""
    if current_test_info is None:
      self._capture_file_path = None
      self._remove_capture_dir()
      return

    dest_dir = current_test_info.output_path
    self._post_process_packet_files(dest_dir)
    self._remove_capture_dir()

  def _post_process_packet_files(self, dest_dir: str):
    """Merges multiple packet files and move the packet file to the dest dir."""
    # Print the capture files stats in logs.
    wifi_utils.run_command(f'ls -alh {self._packet_dir}')
    pcap_files = os.listdir(self._packet_dir)
    if not pcap_files:
      self.log.warning(
          'Packet capture file does not exist on device. This might be because'
          ' no packets were captured if the capturing time was too short.'
      )
      return

    pcap_local_paths = [
        os.path.join(self._packet_dir, name) for name in pcap_files
    ]

    # Skips merge if `mergecap` is not installed.
    which_mergecap = utils.run_command(['which', 'mergecap'])
    if which_mergecap[0] != 0:
      self.log.error(
          'Please install mergecap on your host machine. Otherwise, we will'
          ' directly upload all capture files without merging them into one'
          ' capture file.'
      )
      for path in pcap_local_paths:
        # If mergecap is not installed, we can set `_capture_file_path` to any
        # of the capture file path.
        self._capture_file_path = shutil.move(path, dest_dir)
      return

    new_file_path = self._merge_capture_files(pcap_local_paths, dest_dir)
    self._capture_file_path = new_file_path

  def _merge_capture_files(self, pcap_local_paths: list[str], dest_dir: str):
    filename = os.path.splitext(os.path.basename(pcap_local_paths[0]))[0]
    new_file_name = filename.replace(_FILE_TAG, _FILE_TAG_MERGED)
    new_file_name = f'{new_file_name}.pcap'
    new_file_path = os.path.join(dest_dir, new_file_name)
    self.log.debug(
        'Merging following capture files into one file %s: %s',
        os.path.basename(new_file_path),
        ','.join(pcap_local_paths),
    )
    result = utils.run_command(
        ['mergecap', '-w', new_file_path] + pcap_local_paths
    )
    if result[0] != 0:
      raise errors.SnifferManagerError(
          f'Failed to merge capture files with result: {result}'
      )
    for pcap_local_path in pcap_local_paths:
      os.remove(pcap_local_path)
    return new_file_path

  def _get_sniffer_log_path(self, log_level: str):
    return os.path.join(
        self.device.log_path, f'{_FILE_TAG},{self._timestamp}.{log_level}'
    )

  def _get_packet_streaming_path(self):
    return os.path.join(self._packet_dir, f'{_FILE_TAG},{self._timestamp}.pcap')



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Installs Mobly WiFi Module."""

import setuptools

description = (
    'Mobly WiFi controller module for using Python code to operate network '
    'devices in Mobly tests.'
)

install_requires = [
    'mobly>=1.12.2',
    'immutabledict',
    'paramiko>=2.10.4',
    'pyzmq>=15.0.0',
    'packaging',
]

setuptools.setup(
    name='mobly-wifi',
    version='1.2.0',
    author='Minghao Li',
    author_email='minghaoli@google.com',
    description=description,
    license='Apache2.0',
    url='https://github.com/google/mobly-wifi',
    packages=setuptools.find_namespace_packages(
        include=['mobly.controllers.*']
    ),
    package_data={'mobly.controllers.wifi': ['data/*']},
    install_requires=install_requires,
    python_requires='>=3.11',
)
